@@ -28,7 +28,7 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 #include <assert.h>
 
-/* 
+/*
  * This file contains macros to manipulate singly and doubly-linked lists.
  *
  * 1. LL_ macros:  singly-linked lists.
@@ -38,7 +38,7 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  * To use singly-linked lists, your structure must have a "next" pointer.
  * To use doubly-linked lists, your structure must "prev" and "next" pointers.
  * Either way, the pointer to the head of the list must be initialized to NULL.
- * 
+ *
  * ----------------.EXAMPLE -------------------------
  * struct item {
  *      int id;
@@ -64,14 +64,16 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    when compiling c++ code), this code uses whatever method is needed
    or, for VS2008 where neither is available, uses casting workarounds. */
 #ifdef _MSC_VER            /* MS compiler */
-#if _MSC_VER >= 1600 && defined(__cplusplus)  /* VS2010 or newer in C++ mode */
-#define LDECLTYPE(x) decltype(x)
-#else                     /* VS2008 or older (or VS2010 in C mode) */
-#define NO_DECLTYPE
-#define LDECLTYPE(x) char*
-#endif
+#  if _MSC_VER >= 1600 && defined(__cplusplus) /* VS2010 or newer in C++ mode */
+#    define LDECLTYPE(x) decltype(x)
+#  elif _MSC_VER >= 1928   /* VS2019 v16.8 for C and C++ mode */
+#    define LDECLTYPE(x) __typeof__(x)
+#  else                    /* VS2008 or older (or VS2010 in C mode) */
+#    define NO_DECLTYPE
+#    define LDECLTYPE(x) char*
+#  endif
 #else                      /* GNU, Sun and other compilers */
-#define LDECLTYPE(x) __typeof(x)
+#  define LDECLTYPE(x) __typeof(x)
 #endif
 
 /* for VS2008 we use some workarounds to get around the lack of decltype,
@@ -85,7 +87,7 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #define _PREVASGN(elt,list,to) { char **_alias = (char**)&((list)->prev); *_alias=(char*)(to); }
 #define _RS(list) { char **_alias = (char**)&(list); *_alias=_tmp; }
 #define _CASTASGN(a,b) { char **_alias = (char**)&(a); *_alias=(char*)(b); }
-#else 
+#else
 #define _SV(elt,list)
 #define _NEXT(elt,list) ((elt)->next)
 #define _NEXTASGN(elt,list,to) ((elt)->next)=(to)
@@ -380,19 +382,27 @@ do {
 #define LL_FOREACH_SAFE(head,el,tmp)                                                           \
   for((el)=(head);(el) && (tmp = (el)->next, 1); (el) = tmp)
 
-#define LL_SEARCH_SCALAR(head,out,field,val)                                                   \
-do {                                                                                           \
-    LL_FOREACH(head,out) {                                                                     \
-      if ((out)->field == (val)) break;                                                        \
-    }                                                                                          \
-} while(0) 
+#define LL_SEARCH_SCALAR(head, out, field, val)                               \
+  do                                                                          \
+    {                                                                         \
+      LL_FOREACH (head, out)                                                  \
+      {                                                                       \
+        if ((out)->field == (val))                                            \
+          break;                                                              \
+      }                                                                       \
+    }                                                                         \
+  while (0)
 
-#define LL_SEARCH(head,out,elt,cmp)                                                            \
-do {                                                                                           \
-    LL_FOREACH(head,out) {                                                                     \
-      if ((cmp(out,elt))==0) break;                                                            \
-    }                                                                                          \
-} while(0) 
+#define LL_SEARCH(head, out, elt, cmp)                                        \
+  do                                                                          \
+    {                                                                         \
+      LL_FOREACH (head, out)                                                  \
+      {                                                                       \
+        if ((cmp (out, elt)) == 0)                                            \
+          break;                                                              \
+      }                                                                       \
+    }                                                                         \
+  while (0)
 
 /******************************************************************************
  * doubly linked list macros (non-circular)                                   *
@@ -496,27 +506,34 @@ do {
   }                                                                                            \
 } while (0);
 
-#define CDL_FOREACH(head,el)                                                                   \
-    for(el=head;el;el=(el->next==head ? 0L : el->next)) 
+#define CDL_FOREACH(head, el)                                                 \
+  for (el = head; el; el = (el->next == head ? 0L : el->next))
 
 #define CDL_FOREACH_SAFE(head,el,tmp1,tmp2)                                                    \
   for((el)=(head), ((tmp1)=(head)?((head)->prev):NULL);                                        \
       (el) && ((tmp2)=(el)->next, 1);                                                          \
       ((el) = (((el)==(tmp1)) ? 0L : (tmp2))))
 
-#define CDL_SEARCH_SCALAR(head,out,field,val)                                                  \
-do {                                                                                           \
-    CDL_FOREACH(head,out) {                                                                    \
-      if ((out)->field == (val)) break;                                                        \
-    }                                                                                          \
-} while(0) 
+#define CDL_SEARCH_SCALAR(head, out, field, val)                              \
+  do                                                                          \
+    {                                                                         \
+      CDL_FOREACH (head, out)                                                 \
+      {                                                                       \
+        if ((out)->field == (val))                                            \
+          break;                                                              \
+      }                                                                       \
+    }                                                                         \
+  while (0)
 
-#define CDL_SEARCH(head,out,elt,cmp)                                                           \
-do {                                                                                           \
-    CDL_FOREACH(head,out) {                                                                    \
-      if ((cmp(out,elt))==0) break;                                                            \
-    }                                                                                          \
-} while(0) 
+#define CDL_SEARCH(head, out, elt, cmp)                                       \
+  do                                                                          \
+    {                                                                         \
+      CDL_FOREACH (head, out)                                                 \
+      {                                                                       \
+        if ((cmp (out, elt)) == 0)                                            \
+          break;                                                              \
+      }                                                                       \
+    }                                                                         \
+  while (0)
 
 #endif /* UTLIST_H */
-
