@@ -41,25 +41,25 @@ IGNORE_COMPILER_WARNING("-Wstrict-aliasing")
 #include <clang/Frontend/TextDiagnosticBuffer.h>
 #include <clang/Frontend/TextDiagnosticPrinter.h>
 
-#ifdef LLVM_OLDER_THAN_10_0
-#include "llvm/ADT/ArrayRef.h"
-#endif
+#include <llvm/IR/Function.h>
+#include <llvm/IR/Module.h>
+#include <llvm/IR/Verifier.h>
+#include <llvm/LinkAllPasses.h>
+#include <llvm/Linker/Linker.h>
+#include <llvm/Support/MemoryBuffer.h>
+#include <llvm/Support/VirtualFileSystem.h>
+#include <llvm/Transforms/Utils/Cloning.h>
 
-#include "llvm/LinkAllPasses.h"
-#include "llvm/Linker/Linker.h"
-
-#include "llvm/Transforms/Utils/Cloning.h"
-
-#include "llvm/IR/Function.h"
-#include "llvm/IR/Module.h"
-
-#ifndef LLVM_OLDER_THAN_11_0
+#if LLVM_VERSION_MAJOR > 15
+#include "llvm/TargetParser/Host.h"
+#elif LLVM_VERSION_MAJOR > 10
 #include "llvm/Support/Host.h"
 #endif
 
 #include <iostream>
-#include <sstream>
+#include <map>
 #include <regex>
+#include <sstream>
 
 // For some reason including pocl.h before including CodeGenAction.h
 // causes an error. Some kind of macro definition issue. To investigate.
@@ -68,18 +68,28 @@ IGNORE_COMPILER_WARNING("-Wstrict-aliasing")
 // causing compilation error if they are included before the LLVM headers.
 #include "pocl_llvm_api.h"
 #include "pocl_runtime_config.h"
-#include "linker.h"
 #include "pocl_file_util.h"
 #include "pocl_cache.h"
 #include "LLVMUtils.h"
 #include "pocl_util.h"
 
+#ifdef ENABLE_HEADER_BUNDLING
+#include "HeaderBundle/_clang_opencl.h"
+#include "HeaderBundle/opencl-c-base.h"
+#include "HeaderBundle/opencl-c.h"
+//
+// TODO it should be possible to pass include files to Clang entirely in memory
+// (via llvm::MemoryBuffer), however this doesn't seem to not work
+// #define BUNDLED_HEADERS_IN_MEMORY
+#endif
+
 using namespace clang;
 using namespace llvm;
 
 POP_COMPILER_DIAGS
 
-
+#include "ProgramScopeVariables.h"
+#include "linker.h"
 
 //#define DEBUG_POCL_LLVM_API
 
@@ -154,19 +164,141 @@ static void get_build_log(cl_program program,
 static llvm::Module *getKernelLibrary(cl_device_id device,
                                       PoclLLVMContextData *llvm_ctx);
 
+/**
+ * Runs various LLVM "passes" on the program.bc LLVM module;
+ * the passes are not real LLVM passes, but perhaps it will make sense
+ * to convert at some point. Note that this should only run passes which
+ * for some reason must be run at program.bc stage rather than parallel.bc
+ *
+ * \param [in] Context the pocl LLVM context
+ * \param [in] Mod is the LLVM module
+ * \param [in] Program the cl_program corresponding to Mod
+ * \param [in] Device the device used for the LLVM passes
+ * \param [in] device_i index into program->devices[] corresponding to Device
+ * \param [out] Log a std::string containing the error/warning log
+ * \returns true if there is an error
+ *
+ */
+static bool generateProgramBC(PoclLLVMContextData *Context, llvm::Module *Mod,
+                             cl_program Program, cl_device_id Device,
+                             unsigned device_i, std::string &Log) {
+
+  llvm::Module *BuiltinLib = getKernelLibrary(Device, Context);
+  if (BuiltinLib == nullptr)
+    return true;
+
+  if (Device->run_program_scope_variables_pass) {
+    size_t TotalGVarBytes = 0;
+    if (runProgramScopeVariablesPass(Mod, Device->global_as_id,
+                                     Device->local_as_id, TotalGVarBytes, Log))
+      return true;
+    Program->global_var_total_size[device_i] = TotalGVarBytes;
+  }
+
+  std::string Opts;
+  if (Program->compiler_options)
+    Opts.assign(Program->compiler_options);
+  bool DebugRequested = (Opts.find("-g") != std::string::npos);
+  if (link(Mod, BuiltinLib, Log, Device, !DebugRequested))
+    return true;
+
+  raw_string_ostream OS(Log);
+  bool BrokenDebugInfo = false;
+  if (pocl_get_bool_option("POCL_LLVM_VERIFY", LLVM_VERIFY_MODULE_DEFAULT)) {
+    if (llvm::verifyModule(*Mod, &OS, &BrokenDebugInfo))
+      return true;
+  }
+
+  if (BrokenDebugInfo)
+    Log.append("Warning: broken DebugInfo detected\n");
+
+  return false;
+}
+
+#ifdef ENABLE_HEADER_BUNDLING
+static std::map<std::string_view, std::string> UnbundledHeaders;
+static bool UnbundledHeadersInitialized = false;
+
+static void addHeader(const char *Source, unsigned SourceLen,
+                      const char *MapKey) {
+
+#ifdef BUNDLED_HEADERS_IN_MEMORY
+  std::string Temp;
+  Temp.assign(Source, SourceLen);
+  UnbundledHeaders.insert(std::make_pair(MapKey, Temp));
+#else
+  char HdrCachePath[POCL_MAX_PATHNAME_LENGTH];
+  pocl_cache_write_header(HdrCachePath, MapKey, (const char *)Source,
+                          SourceLen);
+  UnbundledHeaders.insert(std::make_pair(MapKey, std::string(HdrCachePath)));
+#endif
+}
+
+static void setupUnbundledHeaders() {
+  if (!UnbundledHeadersInitialized) {
+    addHeader(_clang_opencl_h, _clang_opencl_h_len, "_clang_opencl.h");
+    addHeader(opencl_c_base_h, opencl_c_base_h_len, "opencl-c-base.h");
+    addHeader(opencl_c_h, opencl_c_h_len, "opencl-c.h");
+    UnbundledHeadersInitialized = true;
+  }
+}
+#endif
+
+using MemBufUptr = std::unique_ptr<llvm::MemoryBuffer>;
+
+static void addHeaderInclude(PreprocessorOptions &po, FrontendOptions &fe,
+                             const char *Header, bool UseBundled) {
+#ifdef ENABLE_POCL_BUILDING
+  bool PoclBuilding = pocl_get_bool_option("POCL_BUILDING", 0);
+#else
+  bool PoclBuilding = false;
+#endif
+
+#ifdef ENABLE_HEADER_BUNDLING
+  if (UseBundled && PoclBuilding == false) {
+    auto It = UnbundledHeaders.find(Header);
+    assert(It != UnbundledHeaders.end());
+#ifdef BUNDLED_HEADERS_IN_MEMORY
+    // avoid writing anything to disk, instead use the RemappedFile feature.
+    const std::string &Content = It->second;
+    MemBufUptr MB(llvm::MemoryBuffer::getMemBufferCopy(Content, Header));
+    po.RetainRemappedFileBuffers = true;
+    po.Includes.push_back(Header);
+    // the remapping does not work
+    po.addRemappedFile(Header, MB.release());
+#else
+    const std::string &HdrCachePath = It->second;
+    po.Includes.push_back(HdrCachePath);
+#endif
+    return;
+  }
+#endif
+
+  std::string Result;
+  if (PoclBuilding) {
+    Result = SRCDIR "/include/";
+  } else {
+    char temp[POCL_MAX_PATHNAME_LENGTH];
+    pocl_get_private_datadir(temp);
+    Result = temp;
+    Result.append("/include/");
+  }
+  Result.append(Header);
+  po.Includes.push_back(Result);
+}
+
 int pocl_llvm_build_program(cl_program program,
                             unsigned device_i,
                             cl_uint num_input_headers,
                             const cl_program *input_headers,
                             const char **header_include_names,
                             int linking_program)
-
 {
-  char tempfile[POCL_FILENAME_LENGTH];
-  char program_bc_path[POCL_FILENAME_LENGTH];
+  char tempfile[POCL_MAX_PATHNAME_LENGTH];
+  char program_bc_path[POCL_MAX_PATHNAME_LENGTH];
   tempfile[0] = 0;
   llvm::Module *mod = nullptr;
-  char temp_include_dir[POCL_FILENAME_LENGTH];
+  char temp_include_dir[POCL_MAX_PATHNAME_LENGTH];
   std::string user_options(program->compiler_options ? program->compiler_options
                                                      : "");
   size_t n = 0;
@@ -199,7 +331,7 @@ int pocl_llvm_build_program(cl_program program,
         std::string dir(path, 0, (tempdir.size() + 1 + last_slash));
         pocl_mkdir_p(dir.c_str());
       }
-      pocl_write_file(path.c_str(), input_header, input_header_size, 0, 1);
+      pocl_write_file(path.c_str(), input_header, input_header_size, 0);
     }
   }
   // Use CompilerInvocation::CreateFromArgs to initialize
@@ -248,8 +380,10 @@ int pocl_llvm_build_program(cl_program program,
   if (device->has_64bit_long)
     ss << "-Dcl_khr_int64 ";
 
-  ss << "-DPOCL_DEVICE_ADDRESS_BITS=" << device->address_bits << " ";
-  ss << "-D__USE_CLANG_OPENCL_C_H ";
+  if (!device->use_only_clang_opencl_headers) {
+    ss << "-DPOCL_DEVICE_ADDRESS_BITS=" << device->address_bits << " ";
+    ss << "-D__USE_CLANG_OPENCL_C_H ";
+  }
 
   ss << "-xcl ";
   // Remove the inline keywords to force the user functions
@@ -261,9 +395,7 @@ int pocl_llvm_build_program(cl_program program,
   // required for clGetKernelArgInfo()
   ss << "-cl-kernel-arg-info ";
 
-  size_t fastmath_flag = user_options.find("-cl-fast-relaxed-math");
-
-#if (CLANG_MAJOR == 15)
+#if (LLVM_MAJOR == 15) || (LLVM_MAJOR == 16)
 #ifdef LLVM_OPAQUE_POINTERS
   ss << "-opaque-pointers ";
 #else
@@ -271,25 +403,35 @@ int pocl_llvm_build_program(cl_program program,
 #endif
 #endif
 
+  std::string fp_contract;
+  if (device->llvm_fp_contract_mode != NULL) {
+    fp_contract = std::string(device->llvm_fp_contract_mode);
+  } else {
+    fp_contract = "on";
+  }
+
+  size_t fastmath_flag = user_options.find("-cl-fast-relaxed-math");
+
   if (fastmath_flag != std::string::npos) {
 #ifdef ENABLE_CONFORMANCE
     user_options.replace(fastmath_flag, 21,
                          "-cl-finite-math-only -cl-unsafe-math-optimizations");
 #endif
     ss << "-D__FAST_RELAXED_MATH__=1 ";
+    fp_contract = "fast";
   }
 
-#ifdef ENABLE_CONFORMANCE
   size_t unsafemath_flag = user_options.find("-cl-unsafe-math-optimizations");
 
   if (unsafemath_flag != std::string::npos) {
+#ifdef ENABLE_CONFORMANCE
     // this should be almost the same but disables -freciprocal-math.
     // required for conformance_math_divide test to pass with OpenCL 3.0
     user_options.replace(unsafemath_flag, 29,
                          "-cl-no-signed-zeros -cl-mad-enable -ffp-contract=fast");
-  }
 #endif
-
+    fp_contract = "fast";
+  }
 
   ss << user_options << " ";
 
@@ -298,10 +440,24 @@ int pocl_llvm_build_program(cl_program program,
 
   if (device->image_support)
     ss << "-D__IMAGE_SUPPORT__=1 ";
+  else {
+    // workaround for a bug in Clang. It unconditionally predefines this macro
+    // when compiling for SPIR or SPIRV target
+    ss << "-U__IMAGE_SUPPORT__ ";
+    ss << "-U__opencl_c_images ";
+    ss << "-U__opencl_c_read_write_images ";
+    ss << "-U__opencl_c_3d_image_writes ";
+    // required for SPIR-V
+    ss << "-D__undef___opencl_c_read_write_images ";
+  }
+  if (device->wg_collective_func_support == CL_FALSE)
+    ss << "-D__undef___opencl_c_work_group_collective_functions ";
+  if ((device->atomic_memory_capabilities &
+       CL_DEVICE_ATOMIC_SCOPE_ALL_DEVICES) == 0)
+    ss << "-D__undef___opencl_c_atomic_scope_all_devices ";
 
   ss << "-DCL_DEVICE_MAX_GLOBAL_VARIABLE_SIZE=" << device->global_var_max_size << " ";
 
-
   ss << "-D__OPENCL_VERSION__=" << device->version_as_int << " ";
 
   if (user_options.find("-cl-std=") == std::string::npos)
@@ -316,7 +472,7 @@ int pocl_llvm_build_program(cl_program program,
   ss << "-D__OPENCL_C_VERSION__=" << cl_std_i << " ";
 
   std::string exts = device->extensions;
-  if (cl_std_major >= 3) {
+  if (cl_std_major >= 3 && device->features != nullptr) {
     exts += ' ';
     exts += device->features;
   }
@@ -331,6 +487,8 @@ int pocl_llvm_build_program(cl_program program,
       if (e_start >= extensions.size())
         break;
       e_end = extensions.find(' ', e_start);
+      if (e_end > extensions.size())
+        e_end = extensions.size();
       llvm::StringRef tok = extensions.slice(e_start, e_end);
       e_start = e_end + 1;
       ss << "-D" << tok.str() << "=1 ";
@@ -338,38 +496,40 @@ int pocl_llvm_build_program(cl_program program,
       cl_ext += tok.str();
       cl_ext += ",";
     }
+    if (device->image_support == CL_FALSE)  {
+      cl_ext += "-__opencl_c_images,";
+      cl_ext += "-__opencl_c_read_write_images,";
+      cl_ext += "-__opencl_c_3d_image_writes,";
+    }
   }
   if (!cl_ext.empty()) {
     cl_ext.back() = ' '; // replace last "," with space
     ss << "-cl-ext=-all," << cl_ext;
   }
 
+  // do not use LLVM builtin functions, rely on PoCL bitcode library only
   ss << "-fno-builtin ";
-  /* with fp-contract=on we get calls to fma with processors which do not
-   * have fma instructions. These ruin the performance.
-   *
-   * TODO find out which processors. Seems to be at least TCE
-   *
-   * default fp-contract is "on" which means "enable if enabled by a pragma".
-   */
-  llvm::Triple triple (device->llvm_target_triplet);
-  if (triple.getArch () == Triple::tce)
-    ss << "-ffp-contract=off ";
+  // do not use jump/switch tables, these create a problem for VUA pass
+  ss << "-fno-jump-tables ";
 
   // This is required otherwise the initialization fails with
   // unknown triple ''
   ss << "-triple=" << device->llvm_target_triplet << " ";
   if (device->llvm_cpu != NULL)
     ss << "-target-cpu " << device->llvm_cpu << " ";
+  if (device->llvm_abi != NULL)
+    ss << "-target-abi " << device->llvm_abi << " ";
 
-  POCL_MSG_PRINT_LLVM("all build options: %s\n", ss.str().c_str());
+  std::string AllBuildOpts = ss.str();
+
+  POCL_MSG_PRINT_LLVM("all build options: %s\n", AllBuildOpts.c_str());
 
   char WSReplacementChar = 0;
 
-  char *TempOptions = (char *) malloc (ss.str().length() + 1);
+  char *TempOptions = (char *)malloc(AllBuildOpts.length() + 1);
 
-  memset (TempOptions, 0, ss.str().length() + 1);
-  strncpy (TempOptions, ss.str().c_str(), ss.str().length());
+  memset(TempOptions, 0, AllBuildOpts.length() + 1);
+  strncpy(TempOptions, AllBuildOpts.c_str(), AllBuildOpts.length());
 
   if (pocl_escape_quoted_whitespace (TempOptions, &WSReplacementChar) == -1)
   {
@@ -394,12 +554,22 @@ int pocl_llvm_build_program(cl_program program,
       }
     }
 
-    // if quoted, remove it to make compiler happy
-    if (s.find("\"") != std::string::npos)
-    {
-      std::regex Target("\"");
-      std::string Replacement = " ";
-      s = std::regex_replace(s, Target, Replacement);
+    // Remove the quotes, Clang doesn't parse them. Note: cannot
+    // replace with whitespace as Clang (at least v15) gets confused by
+    // the whitespace (perhaps thinks it's part of the dir).
+    s = std::regex_replace(s, std::regex("\""), "");
+
+    // Clang (at least v15) gets confused if there's space after -I and
+    // silently fails to add the include directory. Remove the space.
+    // There can be space even without quotes in the user input.
+    s = std::regex_replace(s, std::regex("-I(\\s+)"), "-I");
+
+    // Convert the -g to more specific debug options understood by CFE.
+    if (s == "-g") {
+      itemstrs.push_back("-debug-info-kind=limited");
+      itemstrs.push_back("-dwarf-version=4");
+      itemstrs.push_back("-debugger-tuning=gdb");
+      continue;
     }
 
     itemstrs.push_back(s);
@@ -414,20 +584,21 @@ int pocl_llvm_build_program(cl_program program,
   POCL_MEM_FREE (TempOptions);
 
 #ifdef DEBUG_POCL_LLVM_API
-  // TODO: for some reason the user_options are replicated,
-  // they appear twice in a row in the output
-  std::cerr << "### options: " << ss.str()
-            << "user_options: " << user_options << std::endl;
+  std::cerr << "### options: " << AllBuildOpts << std::endl
+            << std::endl
+            << "user_options: " << user_options << std::endl
+            << std::endl
+            << "c_strs: ";
+  for (auto cstr : itemcstrs) {
+    std::cerr << cstr << " ";
+  }
+  std::cerr << std::endl;
 #endif
 
   if (!CompilerInvocation::CreateFromArgs(
           pocl_build,
-#ifndef LLVM_OLDER_THAN_10_0
           ArrayRef<const char *>(itemcstrs.data(),
                                  itemcstrs.data() + itemcstrs.size()),
-#else
-          itemcstrs.data(), itemcstrs.data() + itemcstrs.size(),
-#endif
           diags)) {
     pocl_cache_create_program_cachedir(program, device_i, program->source,
                                        strlen(program->source),
@@ -437,25 +608,23 @@ int pocl_llvm_build_program(cl_program program,
     return CL_INVALID_BUILD_OPTIONS;
   }
 
+#if LLVM_MAJOR < 18
   LangOptions *la = pocl_build.getLangOpts();
+#else
+  LangOptions L = pocl_build.getLangOpts();
+  LangOptions *la = &L;
+#endif
   PreprocessorOptions &po = pocl_build.getPreprocessorOpts();
+  llvm::Triple triple (device->llvm_target_triplet);
 
-#ifndef LLVM_OLDER_THAN_15_0
+#if LLVM_MAJOR >= 15
   LangOptions::setLangDefaults(*la, clang::Language::OpenCL, triple,
                                po.Includes, clang::LangStandard::lang_opencl12);
 #else
   pocl_build.setLangDefaults(*la,
-#ifndef LLVM_OLDER_THAN_10_0
                              clang::InputKind(clang::Language::OpenCL),
-#else
-                             clang::InputKind::OpenCL,
-#endif
                              triple,
-#ifndef LLVM_OLDER_THAN_12_0
                              po.Includes,
-#else
-                             po,
-#endif
                              clang::LangStandard::lang_opencl12);
 #endif
 
@@ -469,58 +638,51 @@ int pocl_llvm_build_program(cl_program program,
   la->Blocks = true; //-fblocks
   la->MathErrno = false; // -fno-math-errno
   la->NoBuiltin = true;  // -fno-builtin
+  la->Freestanding = true; // -ffree-standing
   la->AsmBlocks = true;  // -fasm (?)
 
+  // setLangDefaults overrides to FPM_On for OpenCL.
+  // So, we need to manually set it after
+  if (fp_contract == "fast") {
+    la->setDefaultFPContractMode(LangOptions::FPM_Fast);
+  } else if (fp_contract == "on") {
+    la->setDefaultFPContractMode(LangOptions::FPM_On);
+  } else if (fp_contract == "off") {
+    la->setDefaultFPContractMode(LangOptions::FPM_Off);
+  }
+
   la->setStackProtector(LangOptions::StackProtectorMode::SSPOff);
 
   la->PICLevel = PICLevel::BigPIC;
-#ifdef __PPC64__
   la->PIE = 0;
-#else
-  la->PIE = 1;
+
+#ifdef ENABLE_HEADER_BUNDLING
+  setupUnbundledHeaders();
 #endif
 
-  std::string IncludeRoot;
-  std::string KernelH;
-  std::string BuiltinRenamesH;
-  std::string PoclTypesH;
-  std::string ClangResourceDir;
+  FrontendOptions &fe = pocl_build.getFrontendOpts();
+  // The CreateFromArgs created an stdin input which we should remove first.
+  fe.Inputs.clear();
 
-#ifdef ENABLE_POCL_BUILDING
-  if (pocl_get_bool_option("POCL_BUILDING", 0)) {
-    IncludeRoot = SRCDIR;
+  if (device->use_only_clang_opencl_headers) {
+#ifdef ENABLE_HEADER_BUNDLING
+    po.Macros.push_back(
+        std::make_pair(std::string("SKIP_HEADER_INCLUDE"), false));
+    bool UseBundledHeaders = true;
 #else
-  if (0) {
+    bool UseBundledHeaders = false;
 #endif
+    addHeaderInclude(po, fe, "opencl-c-base.h", UseBundledHeaders);
+    addHeaderInclude(po, fe, "opencl-c.h", UseBundledHeaders);
+    addHeaderInclude(po, fe, "_clang_opencl.h", UseBundledHeaders);
   } else {
-    char temp[POCL_FILENAME_LENGTH];
-    pocl_get_private_datadir(temp);
-    IncludeRoot = temp;
-#ifdef ENABLE_RELOCATION
-    ClangResourceDir = IncludeRoot;
-#endif
+    addHeaderInclude(po, fe, "pocl_types.h", false);
+    addHeaderInclude(po, fe, "_builtin_renames.h", false);
+    addHeaderInclude(po, fe, "opencl-c-base.h", false);
+    addHeaderInclude(po, fe, "opencl-c.h", false);
+    addHeaderInclude(po, fe, "_kernel.h", false);
   }
-  if (ClangResourceDir.empty()) {     
-#ifndef LLVM_OLDER_THAN_9_0
-    ClangResourceDir = driver::Driver::GetResourcesPath(CLANG);
-#else
-    DiagnosticsEngine Diags{new DiagnosticIDs, new DiagnosticOptions};
-    driver::Driver TheDriver(CLANG, "", Diags);
-    ClangResourceDir = TheDriver.ResourceDir;
-#endif
-  }
-  KernelH = IncludeRoot + "/include/_kernel.h";
-  BuiltinRenamesH = IncludeRoot + "/include/_builtin_renames.h";
-  PoclTypesH = IncludeRoot + "/include/pocl_types.h";
 
-  po.Includes.push_back(PoclTypesH);
-  po.Includes.push_back(BuiltinRenamesH);
-  // Use Clang's opencl-c.h header.
-#ifndef LLVM_OLDER_THAN_9_0
-  po.Includes.push_back(ClangResourceDir + "/include/opencl-c-base.h");
-#endif
-  po.Includes.push_back(ClangResourceDir + "/include/opencl-c.h");
-  po.Includes.push_back(KernelH);
   clang::TargetOptions &ta = pocl_build.getTargetOpts();
   ta.Triple = device->llvm_target_triplet;
   if (device->llvm_cpu != NULL)
@@ -529,27 +691,23 @@ int pocl_llvm_build_program(cl_program program,
 #ifdef DEBUG_POCL_LLVM_API
   std::cout << "### Triple: " << ta.Triple.c_str() <<  ", CPU: " << ta.CPU.c_str();
 #endif
+#if LLVM_MAJOR < 20
   CI.createDiagnostics(diagsBuffer, false);
-
-  FrontendOptions &fe = pocl_build.getFrontendOpts();
-  // The CreateFromArgs created an stdin input which we should remove first.
-  fe.Inputs.clear();
+#else
+  CI.createDiagnostics(*llvm::vfs::getRealFileSystem(), diagsBuffer, false);
+#endif
 
   // Read input source to clang::FrontendOptions.
   // The source is contained in the program->source array,
   // but if debugging option is enabled in the kernel compiler
   // we need to dump the file to disk first for the debugger
   // to find it.
-  char source_file[POCL_FILENAME_LENGTH];
+  char source_file[POCL_MAX_PATHNAME_LENGTH];
   POCL_RETURN_ERROR_ON(pocl_cache_write_program_source(source_file, program),
                        CL_OUT_OF_HOST_MEMORY, "Could not write program source");
   fe.Inputs.push_back(
       FrontendInputFile(source_file,
-#ifndef LLVM_OLDER_THAN_10_0
                         clang::InputKind(clang::Language::OpenCL)
-#else
-                        clang::InputKind::OpenCL
-#endif
                             ));
 
   CodeGenOptions &cg = pocl_build.getCodeGenOpts();
@@ -627,17 +785,20 @@ int pocl_llvm_build_program(cl_program program,
     program->binary_sizes[device_i] = (size_t)fsize;
     program->binaries[device_i] = (unsigned char *)binary;
 
-    mod = (llvm::Module *)program->data[device_i];
+    mod = (llvm::Module *)program->llvm_irs[device_i];
     if (mod != nullptr) {
       delete mod;
-      program->data[device_i] = nullptr;
+      program->llvm_irs[device_i] = nullptr;
       --llvm_ctx->number_of_IRs;
     }
 
-    program->data[device_i] = parseModuleIR(program_bc_path, llvm_ctx->Context);
-    assert(program->data[device_i]);
+    program->llvm_irs[device_i] = mod =
+        parseModuleIR(program_bc_path, llvm_ctx->Context);
+    assert(mod);
     ++llvm_ctx->number_of_IRs;
 
+    parseModuleGVarSize(program, device_i, mod);
+
     return CL_SUCCESS;
   }
 
@@ -652,9 +813,10 @@ int pocl_llvm_build_program(cl_program program,
   if (!success)
     return CL_BUILD_PROGRAM_FAILURE;
 
-  mod = (llvm::Module *)program->data[device_i];
+  mod = (llvm::Module *)program->llvm_irs[device_i];
   if (mod != nullptr) {
     delete mod;
+    program->llvm_irs[device_i] = nullptr;
     --llvm_ctx->number_of_IRs;
   }
 
@@ -666,20 +828,13 @@ int pocl_llvm_build_program(cl_program program,
 
   if (mod->getModuleFlag("PIC Level") == nullptr)
     mod->setPICLevel(PICLevel::BigPIC);
-#ifndef __PPC64__
-  if (mod->getModuleFlag("PIE Level") == nullptr)
-    mod->setPIELevel(PIELevel::Large);
-#endif
 
   // link w kernel lib, but not if we're called from clCompileProgram()
   // Later this should be replaced with indexed linking of source code
   // and/or bitcode for each kernel.
   if (linking_program) {
-    llvm::Module *libmodule = getKernelLibrary(device, llvm_ctx);
-    assert(libmodule != NULL);
     std::string log("Error(s) while linking: \n");
-    if (link(mod, libmodule, log, device->global_as_id,
-             device->device_aux_functions)) {
+    if (generateProgramBC(llvm_ctx, mod, program, device, device_i, log)) {
       appendToProgramBuildLog(program, device_i, log);
       std::string msg = getDiagString(ctx);
       appendToProgramBuildLog(program, device_i, msg);
@@ -690,12 +845,12 @@ int pocl_llvm_build_program(cl_program program,
     }
   }
 
-  program->data[device_i] = mod;
+  program->llvm_irs[device_i] = mod;
 
   POCL_MSG_PRINT_LLVM("Writing program.bc to %s.\n", program_bc_path);
 
   /* Always retain program.bc */
-  error = pocl_write_module(mod, program_bc_path, 0);
+  error = pocl_write_module(mod, program_bc_path);
   if(error)
     return error;
 
@@ -721,7 +876,6 @@ int pocl_llvm_build_program(cl_program program,
 static int pocl_convert_spir_bitcode_to_target(llvm::Module *p,
                                                llvm::Module *libmodule,
                                                cl_device_id device) {
-#ifdef ENABLE_SPIR
   const std::string &ModTriple = p->getTargetTriple();
   if (ModTriple.find("spir") == 0) {
     POCL_RETURN_ERROR_ON((device->endian_little == CL_FALSE),
@@ -746,16 +900,9 @@ static int pocl_convert_spir_bitcode_to_target(llvm::Module *p,
 
     if (p->getModuleFlag("PIC Level") == nullptr)
       p->setPICLevel(PICLevel::BigPIC);
-#ifndef __PPC64__
-    if (p->getModuleFlag("PIE Level") == nullptr)
-      p->setPIELevel(PIELevel::Large);
-#endif
+    return CL_SUCCESS;
   }
   return CL_SUCCESS;
-#else
-  POCL_MSG_ERR("SPIR not supported\n");
-  return CL_LINK_PROGRAM_FAILURE;
-#endif
 }
 
 int pocl_llvm_link_program(cl_program program, unsigned device_i,
@@ -765,23 +912,28 @@ int pocl_llvm_link_program(cl_program program, unsigned device_i,
                            int link_device_builtin_library,
                            int linking_into_new_cl_program) {
 
-  char program_bc_path[POCL_FILENAME_LENGTH];
+  char program_bc_path[POCL_MAX_PATHNAME_LENGTH];
   std::string concated_binaries;
   size_t n = 0, i;
   cl_device_id device = program->devices[device_i];
-  llvm::Module **modptr = (llvm::Module **)&program->data[device_i];
+  llvm::Module **modptr = (llvm::Module **)&program->llvm_irs[device_i];
   int error;
   cl_context ctx = program->context;
   PoclLLVMContextData *llvm_ctx = (PoclLLVMContextData *)ctx->llvm_context_data;
   PoclCompilerMutexGuard lockHolder(&llvm_ctx->Lock);
 
-  llvm::Module *libmodule = getKernelLibrary(device, llvm_ctx);
-  assert(libmodule != NULL);
+  llvm::Module *LibraryModule = getKernelLibrary(device, llvm_ctx);
+  if (LibraryModule == nullptr)
+    return CL_LINK_PROGRAM_FAILURE;
 
   std::unique_ptr<llvm::Module> mod(
       new llvm::Module(StringRef("linked_program"), *llvm_ctx->Context));
   llvm::Module *LinkedModule = nullptr;
   std::unique_ptr<llvm::Module> TempModule;
+  mod->setTargetTriple(LibraryModule->getTargetTriple());
+  mod->setDataLayout(LibraryModule->getDataLayout());
+  mod->setPICLevel(PICLevel::BigPIC);
+
 
   // link the provided modules together into a single module
   for (i = 0; i < num_input_programs; i++) {
@@ -792,24 +944,20 @@ int pocl_llvm_link_program(cl_program program, unsigned device_i,
 
     if (cur_llvm_irs && cur_llvm_irs[i]) {
       llvm::Module *Ptr = (llvm::Module *)cur_llvm_irs[i];
-#ifdef LLVM_OLDER_THAN_7_0
-      TempModule = llvm::CloneModule(Ptr);
-#else
       TempModule = llvm::CloneModule(*Ptr);
-#endif
     } else {
       llvm::Module *Ptr =
-          parseModuleIRMem((char *)cur_device_binaries[0],
-                           cur_device_binary_sizes[0], llvm_ctx->Context);
+          parseModuleIRMem((char *)cur_device_binaries[i],
+                           cur_device_binary_sizes[i], llvm_ctx->Context);
       POCL_RETURN_ERROR_ON((Ptr == nullptr), CL_LINK_PROGRAM_FAILURE,
                            "could not parse module\n");
       TempModule.reset(Ptr);
     }
 
-    error = pocl_convert_spir_bitcode_to_target(TempModule.get(), libmodule,
+    error = pocl_convert_spir_bitcode_to_target(TempModule.get(), LibraryModule,
                                                 device);
     POCL_RETURN_ERROR_ON((error != CL_SUCCESS), CL_LINK_PROGRAM_FAILURE,
-                         "could connvert SPIR to Target\n");
+                         "could not convert SPIR to Target\n");
 
     if (Linker::linkModules(*mod, std::move(TempModule))) {
       std::string msg = getDiagString(ctx);
@@ -832,8 +980,8 @@ int pocl_llvm_link_program(cl_program program, unsigned device_i,
   if (link_device_builtin_library) {
     // linked all the programs together, now link in the kernel library
     std::string log("Error(s) while linking: \n");
-    if (link(LinkedModule, libmodule, log, device->global_as_id,
-             device->device_aux_functions)) {
+    if (generateProgramBC(llvm_ctx, LinkedModule, program, device, device_i,
+                          log)) {
       appendToProgramBuildLog(program, device_i, log);
       std::string msg = getDiagString(ctx);
       appendToProgramBuildLog(program, device_i, msg);
@@ -870,7 +1018,7 @@ int pocl_llvm_link_program(cl_program program, unsigned device_i,
   POCL_MSG_PRINT_LLVM("Writing program.bc to %s.\n", program_bc_path);
 
   /* Always retain program.bc for metadata */
-  error = pocl_write_module(LinkedModule, program_bc_path, 0);
+  error = pocl_write_module(LinkedModule, program_bc_path);
   if (error)
     return error;
 
@@ -889,46 +1037,6 @@ int pocl_llvm_link_program(cl_program program, unsigned device_i,
   return CL_SUCCESS;
 }
 
-/* for "distro" style kernel libs, return which kernellib to use, at runtime */
-#ifdef KERNELLIB_HOST_DISTRO_VARIANTS
-const char *getX86KernelLibName() {
-  StringMap<bool> Features;
-  const char *res = NULL;
-
-  if (!llvm::sys::getHostCPUFeatures(Features)) {
-    POCL_MSG_WARN ("getX86KernelLibName(): LLVM can't get host CPU flags!\n");
-    /* getX86KernelLibName should only ever be enabled
-       on x86-64, which always has sse2 */
-    return "sse2";
-  }
-
-  if (Features["sse2"])
-    res = "sse2";
-  else
-    POCL_ABORT("Pocl on x86_64 requires at least SSE2\n");
-  if (Features["ssse3"] && Features["cx16"])
-    res = "ssse3";
-  if (Features["sse4.1"] && Features["cx16"])
-    res = "sse41";
-  if (Features["avx"] && Features["cx16"] && Features["popcnt"])
-    res = "avx";
-  if (Features["avx"] && Features["cx16"] && Features["popcnt"] && Features["f16c"])
-    res = "avx_f16c";
-  if (Features["avx"] && Features["cx16"] && Features["popcnt"]
-      && Features["xop"] && Features["fma4"])
-    res = "avx_fma4";
-  if (Features["avx"] && Features["avx2"] && Features["cx16"]
-      && Features["popcnt"] && Features["lzcnt"] && Features["f16c"]
-      && Features["fma"] && Features["bmi"] && Features["bmi2"])
-    res = "avx2";
-  if (Features["avx512f"] )
-    res = "avx512";
-
-  return res;
-}
-#endif
-
-
 /**
  * Return the OpenCL C built-in function library bitcode
  * for the given device.
@@ -942,95 +1050,53 @@ static llvm::Module *getKernelLibrary(cl_device_id device,
   if (kernelLibraryMap->find(device) != kernelLibraryMap->end())
     return kernelLibraryMap->at(device);
 
-  const char *subdir = "host";
-  bool is_host = true;
-  // TODO: move this to the device layer, a property to ask for
-  // the kernel builtin bitcode library name, including its subdir
-#if defined(TCE_AVAILABLE)
-  if (triple.getArch() == Triple::tce || triple.getArch() == Triple::tcele) {
-    subdir = "tce";
-    is_host = false;
-  }
-#endif
-#ifdef BUILD_HSA
-  if (triple.getArch() == Triple::hsail64) {
-    subdir = "hsail64";
-    is_host = false;
-  }
-#endif
-#ifdef AMDGCN_ENABLED
-  if (triple.getArch == Triple::amdgcn) {
-    subdir = "amdgcn";
-    is_host = false;
-  }
-#endif
-#ifdef BUILD_CUDA
-  if (triple.getArch() == Triple::nvptx ||
-      triple.getArch() == Triple::nvptx64) {
-    subdir = "cuda";
-    is_host = false;
-  }
-#endif
-
-  std::string kernellib;
-  std::string kernellib_fallback;
+  std::string BuiltinLibraryCommon, BuiltinLibrary, BuiltinLibraryFallback;
 
 #ifdef ENABLE_POCL_BUILDING
   if (pocl_get_bool_option("POCL_BUILDING", 0)) {
-    kernellib = BUILDDIR;
-    kernellib += "/lib/kernel/";
-    kernellib += subdir;
+    BuiltinLibraryCommon = BUILDDIR;
+    BuiltinLibraryCommon += "/lib/kernel/";
+    BuiltinLibraryCommon += device->kernellib_subdir;
   } else // POCL_BUILDING == 0, use install dir
 #endif
   {
-    char temp[POCL_FILENAME_LENGTH];
+    char temp[POCL_MAX_PATHNAME_LENGTH];
     pocl_get_private_datadir(temp);
-    kernellib = temp;
+    BuiltinLibraryCommon = temp;
   }
 
-  kernellib += "/kernel-";
-  kernellib += device->llvm_target_triplet;
+  BuiltinLibraryCommon += "/";
 
-  if (is_host) {
-    kernellib += '-';
-#ifdef KERNELLIB_HOST_DISTRO_VARIANTS
-    kernellib += getX86KernelLibName();
-#elif defined(HOST_CPU_FORCED)
-    kernellib += OCL_KERNEL_TARGET_CPU;
-#else
-    kernellib_fallback = kernellib;
-    kernellib_fallback += OCL_KERNEL_TARGET_CPU;
-    kernellib_fallback += ".bc";
-    if (device->llvm_cpu)
-      kernellib += device->llvm_cpu;
-#endif
+  BuiltinLibrary = BuiltinLibraryCommon + device->kernellib_name;
+  BuiltinLibrary += ".bc";
+
+  if (device->kernellib_fallback_name) {
+    BuiltinLibraryFallback =
+        BuiltinLibraryCommon + device->kernellib_fallback_name;
+    BuiltinLibraryFallback += ".bc";
   }
-  kernellib += ".bc";
 
-  llvm::Module *lib;
+  llvm::Module *BuiltinLibModule = nullptr;
 
-  if (pocl_exists(kernellib.c_str()))
-    {
-      POCL_MSG_PRINT_LLVM("Using %s as the built-in lib.\n", kernellib.c_str());
-      lib = parseModuleIR(kernellib.c_str(), llvmContext);
-    }
-  else
-    {
-#if !defined(KERNELLIB_HOST_DISTRO_VARIANTS) && !defined(HOST_CPU_FORCED)
-      if (is_host && pocl_exists(kernellib_fallback.c_str()))
-        {
-          POCL_MSG_WARN("Using fallback %s as the built-in lib.\n",
-                        kernellib_fallback.c_str());
-          lib = parseModuleIR(kernellib_fallback.c_str(), llvmContext);
-        }
-      else
-#endif
-        POCL_ABORT("Kernel library file %s doesn't exist.\n", kernellib.c_str());
-    }
-  assert (lib != NULL);
-  kernelLibraryMap->insert(std::make_pair(device, lib));
+  if (pocl_exists(BuiltinLibrary.c_str())) {
+    POCL_MSG_PRINT_LLVM("Using %s as the built-in lib.\n",
+                        BuiltinLibrary.c_str());
+    BuiltinLibModule = parseModuleIR(BuiltinLibrary.c_str(), llvmContext);
+  } else {
+    if (device->kernellib_fallback_name &&
+        pocl_exists(BuiltinLibraryFallback.c_str())) {
+      POCL_MSG_WARN("Using fallback %s as the built-in lib.\n",
+                    BuiltinLibraryFallback.c_str());
+      BuiltinLibModule =
+          parseModuleIR(BuiltinLibraryFallback.c_str(), llvmContext);
+    } else
+      POCL_MSG_ERR("Can't find either kernel library file %s or fallback %s.\n",
+                   BuiltinLibrary.c_str(), BuiltinLibraryFallback.c_str());
+  }
+  if (BuiltinLibModule)
+    kernelLibraryMap->insert(std::make_pair(device, BuiltinLibModule));
 
-  return lib;
+  return BuiltinLibModule;
 }
 
 /**
@@ -1042,7 +1108,7 @@ static llvm::Module *getKernelLibrary(cl_device_id device,
  *             the Clang binary.
  * @return 0 on success, error code otherwise.
  */
-int pocl_invoke_clang(cl_device_id Device, const char** Args) {
+int pocl_invoke_clang(const char* TTriple, const char** Args) {
 
   // Borrowed from driver.cpp (clang driver). We do not really care about
   // diagnostics, but just want to get the compilation command invoked with
@@ -1056,18 +1122,28 @@ int pocl_invoke_clang(cl_device_id Device, const char** Args) {
 
   DiagnosticsEngine Diags(DiagID, &*DiagOpts, DiagClient);
 
-  driver::Driver TheDriver(CLANG, Device->llvm_target_triplet, Diags);
+  clang::driver::Driver TheDriver(pocl_get_path("CLANG", CLANGCC),
+                                  TTriple, Diags);
 
   const char **ArgsEnd = Args;
-  while (*ArgsEnd++ != nullptr) {}
+  while (*ArgsEnd != nullptr) { ArgsEnd++; }
+  llvm::SmallVector<const char*, 0> ArgsArray(Args, ArgsEnd);
 
-  llvm::ArrayRef<const char*> ArgsArray(Args, ArgsEnd);
+  int NumExtraArgs;
+  const char *ExtraArgs = pocl_get_args("CLANG", &NumExtraArgs);
+  const char *ExtraArg = ExtraArgs;
+  for (int i = 0; i < NumExtraArgs; ++i) {
+    ArgsArray.push_back(ExtraArg);
+    ExtraArg += strlen(ExtraArg) + 1;
+  }
 
-  std::unique_ptr<driver::Compilation> C(
-    TheDriver.BuildCompilation(ArgsArray));
+  std::unique_ptr<clang::driver::Compilation> C(
+      TheDriver.BuildCompilation(ArgsArray));
+
+  free((void *)ExtraArgs);
 
   if (C && !C->containsError()) {
-    SmallVector<std::pair<int, const driver::Command *>, 4> FailingCommands;
+    SmallVector<std::pair<int, const clang::driver::Command *>, 4> FailingCommands;
     return TheDriver.ExecuteCompilation(*C, FailingCommands);
   } else {
     return -1;
